' Create a Compare Interrupt using the Special Event mechanism
' Compares the 16-bit value of Timer3 with the contents of 16-bit registers CCPR2L\H
' Triggers the interrupt when Timer3 reaches the value held in CCPR2L\H
' Timer3 is automatically reset in hardware
' The rate of the interrupt can be altered by changing the constant value held in cMicroSeconds
'
' Within the interrupt create a simple PWM routine with variable frequency and duty cycle
' The 8-bit relative frequency of the PWM can be altered by changing the value held in bRelativeFrequency
' The 8-bit duty cycle of the PWM can be altered by changing the value held in bDutyCycle
' The duty cycle is not duly effected by relative frequency
'
    Include "Amicus18.inc"                          ' Configure the compiler to use the Amicus18 board (18F25K20)
    Include "Amicus18_Timers.inc"                   ' Include the Amicus18 Timer macros into the program

    On_Hardware_Interrupt GoTo ISR_CCP2             ' Point the interrupt to its handler subroutine

    Symbol tOutputPin = PORTB.0                     ' The designated pin for PWM output
'    
' Variables
'   
    Dim ISR_bFreqAccum As Byte System               ' Frequency accumulator
    Dim ISR_bPwmAccum As Byte System                ' Duty Cycle accumulator   
    
    Dim bDutyCycle As Byte System                   ' Duty Cycle variable
    Dim bRelativeFrequency As Byte System           ' Frequency variable
    
    Dim wCCPR2 As CCPR2L.Word                       ' Combine CCPR2L\H into a 16-bit register

'--------------------------------------------------------------------------------------------------
    GoTo Main                                       ' Jump over the subroutines
'--------------------------------------------------------------------------------------------------
' Special Even Interrupt Handler
' Create a very slow PWM signal with variable frequency and duty 
'
ISR_CCP2:
    Context Save
    If PIR2bits_CCP2IF = 1 Then                     ' Was it a Compare Special Event on CCP2 that triggered the interrupt?
        Inc ISR_bFreqAccum                          ' Yes. So increase the relative frequency accumulator
        If ISR_bFreqAccum = bRelativeFrequency Then ' Has the relative frequency reached the desired value?
            ISR_bFreqAccum = 0                      ' Yes. So reset the relative frequency accumulator          
            Inc ISR_bPwmAccum                       ' Increment the PWM duty cycle accumulator
            If ISR_bPwmAccum >= bDutyCycle Then     ' Has the duty cycle accumulator reached the required duty cycle?
                Clear tOutputPin                    ' Yes. So make the pin low
            Else                                    ' Otherwise...
                Set tOutputPin                      ' Make the pin high
            EndIf
        EndIf
        PIR2bits_CCP2IF = 0                         ' Clear the interrupt flag
    EndIf
'
' << More Interrupt Conditions and Code Here >>
'
    Context Restore                                 ' Exit the interrupt
'--------------------------------------------------------------------------------------------------
' Initialise the interrupt and the PWM variables
'
PWM_Init:
    ISR_bFreqAccum = 1                              ' Reset the ISR relative frequency accumulator
    ISR_bPwmAccum = 0                               ' Reset the ISR duty cycle accumulator      
    bDutyCycle = 0                                  ' Reset the duty cycle
    bRelativeFrequency = 0                          ' Reset the relative frequency
    Low tOutputPin                                  ' Make the pin output low
    
    CCP2CON = %00001011                             ' Compare mode: trigger special event, reset timer, start A/D conversion on CCP2 match (CCP2IF bit is set)

    OpenTimer3(TIMER_INT_OFF & T3_16BIT_RW & T3_PS_1_8 & T3_SYNC_EXT_OFF & T3_SOURCE_INT & T3_SOURCE_CCP)
    $define cPrescalerValue 8                       ' Alter this to match the prescaler parameter above. i.e. 4 for T3_PS_1_4, 8 for T3_PS_1_8
    $define cMicroSeconds 50                        ' Interrupt rate (in uS)
'
' Calculate the value to place into the CCPRx register in order to achieve a certain interrupt rate (in us)
'
    $define cCCPR_Value $eval (cMicroSeconds / cPrescalerValue) * (_xtal / 4)

$if cCCPR_Value > 65535
    $error "Value too large for interrupt duration"
$elseif cCCPR_Value = 0
    $error "Value too small for interrupt duration"
$endif

    wCCPR2 = cCCPR_Value                            ' Load CCPR2L\H with the value to trigger an interrupt at a certain duration
    PIE2bits_CCP2IE = 1                             ' Enable the Special Event Interrupt on CCP2
    INTCONbits_PEIE = 1                             ' Enable Peripheral Interrupts
    INTCONbits_GIE = 1                              ' Enable Global Interrupts
    Return
    
'--------------------------------------------------------------------------------------------------
' The main program loop starts here
'
Main:
    GoSub PWM_Init                                  ' Initialise the interrupt
'
' Create a loop to alter the relative frequency of the PWM signal
' The higher the value of bRelativeFrequency, the lower the relative frequency
'
    bDutyCycle = 127                                ' 50% duty cycle
    
    For bRelativeFrequency = 0 To 255               ' Decrease the frequency
        DelayMS 100                                 ' Create a delay slower than the rate of change
    Next
'
' Create a loop to alter the duty cycle of the PWM signal
' The higher the value of bDutyCycle, the higher the duty cycle
'               
    bRelativeFrequency = 10                         ' Choose a relative frequency
    
    While 1 = 1                                     ' Create an infinite loop
        For bDutyCycle = 0 To 255                   ' Increase the duty cycle
            DelayMS 100                             ' Create a delay slower than the rate of change
        Next
    Wend                                            ' Do it forever
